/**
 * SIMPL basic de/serialization.
 *
 * To use this in browser, first compile it into JavaScript using tsc, then
 * browserify it with:
 *
 *   $ browserify --standalone simpl -o ../../bsjsCore/simpl/simplBase.js simplBase.js
 */
/**
 * A set of callback functions, to receive events and take actions while
 * traversing the object tree.
 */
export interface SimplHandlers {
    /**
     * Traversal encounters a scalar element (e.g. a string or a number).
     *
     * @param {any}           elem   The current (scalar) element.
     * @param {Object}        parent The parent element.
     * @param {string|number} key    The key of elem in parent.
     */
    onScalar?(elem: any, parent: Object, key: string | number): void;
    /**
     * Traversal encounters an array element.
     *
     * @param {any[]}         elem   The current (array) element.
     * @param {Object}        parent The parent element.
     * @param {string|number} key    The key of elem in parent.
     */
    onArray?(elem: any[], parent: Object, key: string | number): void;
    /**
     * Traversal finishes an array element.
     *
     * @param {any[]}         elem   The current (array) element.
     * @param {Object}        parent The parent element.
     * @param {string|number} key    The key of elem in parent.
     */
    onArrayEnd?(elem: any[], parent: Object, key: string | number): void;
    /**
     * Traversal finishes with one element in an array and is about to move to the
     * next element.
     *
     * @param {any}    val  The next element to be traversed.
     * @param {any[]}  elem The array.
     * @param {number} i    The index of val in elem.
     */
    onSep?(val: any, elem: any[], i: number): void;
    /**
     * Traversal encounters an object element.
     *
     * @param {Object}        elem   The current (object) element.
     * @param {Object}        parent The parent element.
     * @param {string|number} key    The key of elem in parent.
     */
    onObject?(elem: Object, parent: Object, key: string | number): void;
    /**
     * Before traversing into an object, this callback is called to notify the
     * field names.
     *
     * @param {Object}          elem       The object to be traversed.
     * @param {Object}          parent     The parent object.
     * @param {string|number}   key        The key of elem in parent.
     * @param {string[]}        fieldNames The field names in elem.
     */
    onFieldNames?(elem: Object, parent: Object, name: string | number, fieldNames: string[]): void;
    /**
     * Before traversing into a field, this callback is called to notify the name
     * of that field.
     *
     * @param {any}    field     The field to be traversed.
     * @param {Object} parent    The parent object.
     * @param {string} fieldName The name of field in parent.
     */
    onFieldName?(field: any, parent: Object, fieldName: string): void;
    /**
     * Traversal finishes an object.
     *
     * @param {Object}        elem   The object being finished.
     * @param {Object}        parent The parent object.
     * @param {string|number} key    The key of elem in parent.
     */
    onObjectEnd?(elem: Object, parent: Object, key: string | number): void;
    /**
     * Traversal encounters an object that has been visited before.
     *
     * Traversal should NOT go into visited objects, to prevent infinite loop.
  
     * @param {Object}        elem   The object being revisited.
     * @param {Object}        parent The parent object.
     * @param {string|number} key   The key of elem in parent.
     */
    onObjectRevisit?(elem: Object, parent: Object, key: string | number): void;
    /**
     * Before traversing into a field, this callback is called to notify the name
     * of the field, to determine if the field should be skipped.
     *
     * @param  {any}     field     The field value.
     * @param  {Object}  elem      The element the field belongs to.
     * @param  {string}  fieldName The name of the field.
     * @return {boolean}           Returns true iff the field should be skipped.
     */
    skipField?(field: any, elem: Object, fieldName: string): boolean;
}
export interface SimplOptions {
    /**
     * Show debugging information to console.
     */
    debugging?: boolean;
    /**
     * Make simpl.id always appear before simpl.refs that refer to it in the
     * serialized form.
     *
     * This is required for some simpl implementation to work.
     *
     * Up to 3x slower than JSON.parse() on V8.
     */
    id_before_ref?: boolean;
}
export declare const SIMPL_ID = "simpl.id";
export declare const SIMPL_REF = "simpl.ref";
export declare const SIMPL_VISITED_ID = "_simpl._visited";
/**
 * (For internal use only) Depth-first search on the given object.
 *
 * @param {Object}        obj      The object to be searched.
 * @param {SimplOptions}  options  Options for the DFS process.
 * @param {SimplHandlers} handlers Handlers for events in the DFS process.
 */
export declare function dfs(obj: Object, options?: SimplOptions, handlers?: SimplHandlers): void;
/**
 * Replaces cyclic references with a stub containing only a simpl.ref.
 *
 * Operates on the object in place.
 *
 * @param  {Object}       obj     The object to be collapsed.
 * @param  {SimplOptions} options Options.
 * @return {Object}               Collapsed object.
 */
export declare function graphCollapse(obj: Object, options?: SimplOptions): Object;
/**
 * Replaces each stub containing simpl.ref with the real object identified by
 * simpl.id.
 *
 * Operates on the object in place.
 *
 * @param  {Object}       obj     The object to be expanded.
 * @param  {SimplOptions} options Options.
 * @return {Object}               The expanded object.
 */
export declare function graphExpand(obj: Object, options?: SimplOptions): Object;
/**
 * Serialize an object. The object can contain cyclic references.
 *
 * The input object is modified during the process, but restored before it
 * returns.
 *
 * @param  {Object}       obj     The object to be serialized.
 * @param  {SimplOptions} options Options.
 * @return {string}               The serialized representation.
 */
export declare function serialize(obj: Object, options?: SimplOptions): string;
/**
 * Deserialize a string into an object.
 *
 * @param  {string}       serial  The serialized representation.
 * @param  {SimplOptions} options Options.
 * @return {Object}               The deserialized object.
 */
export declare function deserialize(serial: string, options?: SimplOptions): Object;
